from PySide6 import QtWidgets


class ListItem(QtWidgets.QListWidgetItem):
    def __init__(self, entry, name="", computation=None, parent=None, origin_uuid=None):
        super().__init__()

        self.entry = entry
        self.name = name
        self.computation = computation
        self.origin_uuid = origin_uuid



#!/usr/bin/env python
"""
Base Module for Testing GUI
 - responsible to set up Qt in order to run on multiple platforms
 - responsible to set up Application
class TestBase
    - responsible to set up and tearDown test workspace
    - responsible to create easy access to 'resource' directory
    - responsible to set up ErrorDialog for easy evaluation of raised exceptions
    - provide method `manual_use` created to help test development process
class DragAndDrop
    - responsible to perform Drag and Drop operations
     from source widget - specific point, to destination widget - specific point
"""

from collections.abc import Iterable
import os
import pathlib
import shutil
import sys
import time
import unittest
from unittest import mock

from h5py import File as HDF5
import pyqtgraph
from PySide6 import QtCore, QtGui, QtTest, QtWidgets

from asammdf import mdf
from asammdf.gui.utils import excepthook

if sys.platform == "win32":
    os.environ["QT_QPA_PLATFORM"] = "windows"
elif sys.platform == "linux":
    os.environ["QT_QPA_PLATFORM"] = "xcb"
elif sys.platform == "darwin":
    os.environ["QT_QPA_PLATFORM"] = "cocoa"
else:
    os.environ["QT_QPA_PLATFORM"] = "windows"

app = pyqtgraph.mkQApp()
app.setOrganizationName("py-asammdf")
app.setOrganizationDomain("py-asammdf")
app.setApplicationName("py-asammdf")


@unittest.skipIf(sys.platform == "darwin", "Test Development on MacOS was not done yet.")
class TestBase(unittest.TestCase):
    """
    - setUp and tearDown test workspace
    - provide method to execute widget
    - setUp ErrorDialog for evaluation raised exceptions
    """

    pyqtgraph.setConfigOption("background", "k")
    pyqtgraph.setConfigOption("foreground", "w")

    settings = QtCore.QSettings()
    settings.setValue("zoom_x_center_on_cursor", True)
    settings.setValue("plot_cursor_precision", 6)

    longMessage = False

    resource = os.path.normpath(os.path.join(os.path.dirname(__file__), "resources"))
    test_workspace = os.path.join(os.path.dirname(__file__), "test_workspace")
    screenshots = os.path.join(os.path.dirname(__file__).split("test")[0], "screenshots")

    patchers = []
    # MockClass ErrorDialog
    mc_ErrorDialog = None

    def shortDescription(self):
        return self._testMethodDoc

    @staticmethod
    def manual_use(w, duration=None):
        """
        Execute Widget for debug/development purpose.

        Parameters
        ----------
        duration : float | None
            duration in seconds
        """
        if duration is None:
            duration = 3600
        else:
            duration = abs(duration)

        w.showMaximized()

        loop = QtCore.QEventLoop()
        QtCore.QTimer.singleShot(int(duration * 1000), loop.quit)
        loop.exec_()

        w.showMaximized()

    @staticmethod
    def processEvents(timeout=0.001):
        QtCore.QCoreApplication.processEvents()
        QtCore.QCoreApplication.sendPostedEvents()
        QtCore.QEventLoop.processEvents(QtCore.QEventLoop())
        if timeout:
            time.sleep(timeout)
            QtCore.QCoreApplication.processEvents()
            QtCore.QCoreApplication.sendPostedEvents()

    def setUp(self) -> None:
        if os.path.exists(self.test_workspace):
            try:
                shutil.rmtree(self.test_workspace)
            except PermissionError as e:
                print(e)
        if not os.path.exists(self.screenshots):
            os.makedirs(self.screenshots)

        os.makedirs(self.test_workspace)
        self.mc_ErrorDialog.reset_mock()
        self.processEvents()

    @classmethod
    def setUpClass(cls):
        sys.excepthook = excepthook
        for patcher in (mock.patch("asammdf.gui.utils.ErrorDialog"),):
            _ = patcher.start()
            cls.patchers.append(patcher)
        cls.mc_ErrorDialog = _

    @classmethod
    def tearDownClass(cls):
        for patcher in cls.patchers:
            patcher.stop()

    def tearDown(self):
        self.processEvents()
        path_ = os.path.join(self.screenshots, self.id().split("gui.")[-1].rsplit(".", 1)[0])
        if not os.path.exists(path_):
            os.makedirs(path_)

        w = getattr(self, "widget", None)
        if w:
            w.grab().save(os.path.join(path_, f"{self.id().split('.')[-1]}.png"))
            self.destroy(w)

        self.mc_ErrorDialog.reset_mock()

        if self.test_workspace and pathlib.Path(self.test_workspace).exists():
            try:
                shutil.rmtree(self.test_workspace)
            except PermissionError as e:
                self.destroy(w)
                print(e)

    @staticmethod
    def destroy(w):
        w.close()
        w.destroy()
        w.deleteLater()

    def mouseClick_RadioButton(self, qitem):
        QtTest.QTest.mouseClick(
            qitem,
            QtCore.Qt.MouseButton.LeftButton,
            QtCore.Qt.KeyboardModifier.NoModifier,
            QtCore.QPoint(2, qitem.height() / 2),
        )
        self.processEvents()

    def mouseClick_CheckboxButton(self, qitem):
        # Same function
        self.mouseClick_RadioButton(qitem)

    def mouseClick_WidgetItem(self, qitem):
        if isinstance(qitem, QtWidgets.QTreeWidgetItem):
            widget = qitem.treeWidget()
        elif isinstance(qitem, QtWidgets.QListWidgetItem):
            widget = qitem.listWidget()
        else:
            raise NotImplementedError
        QtTest.QTest.mouseClick(
            widget.viewport(),
            QtCore.Qt.MouseButton.LeftButton,
            QtCore.Qt.KeyboardModifier.NoModifier,
            widget.visualItemRect(qitem).center(),
        )
        self.processEvents(0.5)

    def mouseDClick_WidgetItem(self, qitem):
        if isinstance(qitem, QtWidgets.QTreeWidgetItem):
            widget = qitem.treeWidget()
        elif isinstance(qitem, QtWidgets.QListWidgetItem):
            widget = qitem.listWidget()
        else:
            raise NotImplementedError
        QtTest.QTest.mouseDClick(
            widget.viewport(),
            QtCore.Qt.MouseButton.LeftButton,
            QtCore.Qt.KeyboardModifier.NoModifier,
            widget.visualItemRect(qitem).center(),
        )
        self.processEvents(0.5)

    def avoid_blinking_issue(self, w):
        self.processEvents(0.01)
        # To avoid blinking issue, click on a center of widget
        QtTest.QTest.mouseClick(
            w, QtCore.Qt.MouseButton.LeftButton, QtCore.Qt.KeyboardModifier.NoModifier, w.rect().center()
        )


class DragAndDrop:
    def __init__(self, src_widget, dst_widget, src_pos, dst_pos):
        QtCore.QCoreApplication.processEvents()
        # hack QDrag object
        with mock.patch(f"{src_widget.__module__}.QtGui.QDrag") as mo_QDrag:
            src_widget.startDrag(QtCore.Qt.DropAction.MoveAction)
            mo_QDrag.assert_called()
            mime_data = mo_QDrag.return_value.setMimeData.call_args.args[0]

            event = QtGui.QDragEnterEvent(
                dst_pos,
                QtCore.Qt.DropAction.MoveAction,
                mime_data,
                QtCore.Qt.MouseButton.LeftButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
            )
            with mock.patch.object(event, "source", return_value=src_widget):
                dst_widget.dragEnterEvent(event)

            event = QtGui.QDropEvent(
                dst_pos,
                QtCore.Qt.DropAction.MoveAction,
                mime_data,
                QtCore.Qt.MouseButton.LeftButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
            )
            with mock.patch.object(event, "source", return_value=src_widget):
                dst_widget.dropEvent(event)
        QtCore.QCoreApplication.processEvents()


class Pixmap:
    COLOR_BACKGROUND = "#000000"
    COLOR_RANGE = "#000032"
    COLOR_CURSOR = "#e69138"

    @staticmethod
    def is_black(pixmap) -> bool:
        """
        Excepting cursor
        """
        image = pixmap.toImage()

        for y in range(image.height()):
            for x in range(image.width()):
                color = QtGui.QColor(image.pixel(x, y))
                if color.name() not in (Pixmap.COLOR_BACKGROUND, Pixmap.COLOR_CURSOR):
                    return False
        return True

    @staticmethod
    def is_colored(pixmap, color_name, x, y, width=None, height=None):
        image = pixmap.toImage()

        offset = 1
        y = y + offset

        if not width:
            width = image.width()
        if not height:
            height = image.height()

        for _y in range(offset, image.height()):
            for _x in range(image.width()):
                color = QtGui.QColor(image.pixel(_x, _y))
                if _x < x or _y < y:
                    continue
                # De unde 2?
                elif (_x > width - x) or (_y > height - y - 2):
                    break
                if color.name() != color_name:
                    print(x, y, width, height)
                    print(_x, _y, color.name())
                    return False
        return True

    @staticmethod
    def has_color(pixmap, color_name):
        """
        Return True if "pixmap" has "color_name" color
        """
        image = pixmap.toImage()
        if not isinstance(color_name, str):
            if hasattr(color_name, "color"):
                color_name = color_name.color.name()
            elif hasattr(color_name, "name"):
                color_name = color_name.name()
            else:
                raise SyntaxError(f"Object {color_name} doesn't have the attribute <<color>> or <<name()>>")
        for y in range(image.height()):
            for x in range(image.width()):
                color = QtGui.QColor(image.pixel(x, y))
                if color.name() == color_name:
                    return True
        return False

    @staticmethod
    def color_names(pixmap):
        """

        Parameters
        ----------
        pixmap: QPixmap object of PlotGraphics object

        Returns
        -------
        All colors from pixmap including default colors
        """
        color_names = set()

        image = pixmap.toImage()
        for y in range(image.height()):
            for x in range(image.width()):
                color = QtGui.QColor(image.pixel(x, y))
                color_names.add(color.name())
        return color_names

    @staticmethod
    def color_names_exclude_defaults(pixmap):
        """

        Parameters
        ----------
        pixmap: QPixmap object of PlotGraphics object

        Returns
        -------
        All colors from pixmap excluding default colors
        """
        color_names = set()
        defaults = (Pixmap.COLOR_BACKGROUND, Pixmap.COLOR_CURSOR, Pixmap.COLOR_RANGE)
        image = pixmap.toImage()
        for y in range(image.height()):
            for x in range(image.width()):
                color = QtGui.QColor(image.pixel(x, y))
                if color not in defaults:
                    color_names.add(color.name())
        return color_names

    @staticmethod
    def color_map(pixmap):
        """
        return dict, where:
            > keys are line of pixmap
            > values is a list of color names ordered by columns of pixmap
        """
        color_dict = {}
        line = []
        image = pixmap.toImage()
        for y in range(image.height()):
            for x in range(image.width()):
                line.append(QtGui.QColor(image.pixel(x, y)).name())
            color_dict[y] = line
            line = []
        return color_dict

    @staticmethod
    def cursors_x(pixmap):
        """

        Parameters
        ----------
        pixmap: QPixmap object of PlotGraphics object

        Returns
        -------
        list of cursors line from pixmap
        """
        image = pixmap.toImage()

        cursors = []
        possible_cursor = None

        for x in range(image.width()):
            count = 0
            for y in range(image.height()):
                color = QtGui.QColor(image.pixel(x, y))
                # Count straight vertical line pixels with COLOR_CURSOR color
                if color.name() == Pixmap.COLOR_CURSOR:
                    count += 1
            if count >= (image.height() - 1) / 2 - 1:  # For Y shortcut tests, one cursor is a discontinuous line
                cursors.append(x)
        return cursors

    @staticmethod
    def search_signal_extremes_by_ax(pixmap, signal_color, ax: str):
        """
        Return column where signal start and end
        If ax = Y: Return a list with extremes of signal by 0Y axes
        If ax = X: Return a list with extremes of signal by 0X axes
        """
        if not isinstance(signal_color, str):
            if hasattr(signal_color, "color"):
                signal_color = signal_color.color.name()
            elif hasattr(signal_color, "name"):
                signal_color = signal_color.name()
            else:
                raise SyntaxError(f"Object {signal_color} doesn't have the attribute <<color>> or <<name()>>")
        from_to = []
        image = pixmap.toImage()
        if ax in ("x", "X"):
            for x in range(image.width()):
                for y in range(image.height()):
                    if image.pixelColor(x, y).name() == signal_color:
                        from_to.append(x)
                        break
                if from_to:
                    break
            if not from_to:
                return
            for x in range(image.width() - 1, from_to[0], -1):
                for y in range(image.height()):
                    if image.pixelColor(x, y).name() == signal_color:
                        from_to.append(x)
                        break
                if len(from_to) == 2:
                    break
            return from_to

        elif ax in ("y", "Y"):
            for y in range(image.height()):
                for x in range(image.width()):
                    if image.pixelColor(x, y).name() == signal_color:
                        from_to.append(y)
                        break
                if from_to:
                    break
            if not from_to:
                return

            for y in range(image.height() - 1, from_to[0], -1):
                for x in range(image.width()):
                    if image.pixelColor(x, y).name() == signal_color:
                        from_to.append(y)
                        break
                if len(from_to) == 2:
                    break
            return from_to

    @staticmethod
    def search_y_of_signal_in_column(pixmap_column, signal_color):
        """
        Return the first pixel line number where the signal color was found.
        """
        image = pixmap_column.toImage()
        if image.width() > 1:
            raise TypeError(f"<<{image.width()} != 1>>. Please check pixmap width!")
        if not isinstance(signal_color, str):
            if hasattr(signal_color, "color"):
                signal_color = signal_color.color.name()
            elif hasattr(signal_color, "name"):
                signal_color = signal_color.name()
            else:
                raise SyntaxError(f"Object {signal_color} doesn't have the attribute <<color>> or <<name()>>")

        line = None
        for y in range(image.height()):
            if QtGui.QColor(image.pixel(0, y)).name() == signal_color:
                line = y
                break
        return line


class OpenFileContextManager:
    """
    Generic class for opening a file using context manager.
    Methods:
        __enter__: return opened file object.
        __exit__: close file object. If exc_type, exc_val, exc_tb, raise exception.
    """

    def __init__(self, file_path: str | pathlib.Path):
        """
        Parameters
        ----------
        file_path: file path as str or pathlib.Path object
        """
        self.file = None
        if isinstance(file_path, str):
            self._file_path = pathlib.Path(file_path)
        else:
            self._file_path = file_path

        assert self._file_path.exists(), "Provided file does not exist"

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.close()
        for exc in (exc_type, exc_val, exc_tb):
            if exc is not None:
                print(exc)


class OpenMDF(OpenFileContextManager):
    """
    Open MDF file using context manager.
    """

    def __enter__(self):
        self.file = mdf.MDF(self._file_path, process_bus_logging=("process_bus_logging", True))
        return self.file


class OpenHDF5(OpenFileContextManager):
    """
    Open HDF5 file using context manager.
    """

    def __enter__(self):
        self.file = HDF5(self._file_path)
        return self.file


class DBC:
    class BO:
        def __init__(self, lines: Iterable[str]):
            for line in lines:
                if line.startswith("BO_ "):
                    self.data = line

        def __repr__(self):
            return self.data

        @property
        def name(self) -> str:
            if self.data:
                return self.data.split()[2].strip(":")

        @property
        def id(self) -> str:
            if self.data:
                return self.data.split()[1]

        @property
        def data_length(self) -> int:
            if self.data:
                return int(self.data.split()[3])

    class SG:
        def __init__(self, name: str, lines: Iterable[str]):
            self.name = name
            self.data = None
            for line in lines:
                if "SG_ " in line and name in line:
                    self.data = line.split("SG_ ")[1]

        def __repr__(self):
            return self.data

        @property
        def unit(self) -> str:
            return self.data.split('"')[1]  # middle value, ex: "blah blah "unit" blah

        @property
        def bit_count(self) -> int:
            return int(self.data.split("|")[1].split("@")[0])  # is complicated this CAN :)

        @property
        def conversion_a(self) -> float:
            return float(self.data.split("(")[1].split(")")[0].split(",")[0])

        @property
        def conversion_b(self) -> float:
            return float(self.data.split("(")[1].split(")")[0].split(",")[1])



from pathlib import Path
from tempfile import gettempdir
import threading

from PySide6 import QtGui, QtWidgets

from ...blocks.utils import extract_encryption_information
from ...blocks.v4_constants import FLAG_AT_TO_STRING
from ..ui.attachment import Ui_Attachment

try:
    import sounddevice as sd
    import soundfile as sf

    current_frame = 0
except:
    current_frame = None


class Attachment(Ui_Attachment, QtWidgets.QWidget):
    def __init__(self, index, attachment_block, file, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

        self.extract_btn.clicked.connect(self.extract)
        self.file = file
        self.index = index

        self.number.setText(f"{index + 1}.")

        fields = []

        field = QtWidgets.QTreeWidgetItem()
        field.setText(0, "ATBLOCK address")
        field.setText(1, f"0x{attachment_block.address:X}")
        fields.append(field)

        field = QtWidgets.QTreeWidgetItem()
        field.setText(0, "File name")
        field.setText(1, str(attachment_block.file_name))
        fields.append(field)

        field = QtWidgets.QTreeWidgetItem()
        field.setText(0, "MIME type")
        field.setText(1, attachment_block.mime)
        fields.append(field)

        field = QtWidgets.QTreeWidgetItem()
        field.setText(0, "Comment")
        field.setText(1, attachment_block.comment)
        fields.append(field)

        field = QtWidgets.QTreeWidgetItem()
        field.setText(0, "Flags")
        if attachment_block.flags:
            flags = []
            for flag, string in FLAG_AT_TO_STRING.items():
                if attachment_block.flags & flag:
                    flags.append(string)
            text = f"{attachment_block.flags} [0x{attachment_block.flags:X}= {', '.join(flags)}]"
        else:
            text = "0"
        field.setText(1, text)
        fields.append(field)

        field = QtWidgets.QTreeWidgetItem()
        field.setText(0, "MD5 sum")
        field.setText(1, attachment_block.md5_sum.hex().upper())
        fields.append(field)

        size = attachment_block.original_size
        if size <= 1 << 10:
            text = f"{size} B"
        elif size <= 1 << 20:
            text = f"{size / 1024:.1f} KB"
        elif size <= 1 << 30:
            text = f"{size / 1024 / 1024:.1f} MB"
        else:
            text = f"{size / 1024 / 1024 / 1024:.1f} GB"

        field = QtWidgets.QTreeWidgetItem()
        field.setText(0, "Size")
        field.setText(1, text)
        fields.append(field)

        self.fields.addTopLevelItems(fields)

        self.audio_comment = None
        self.audio_progress = None

        if current_frame is not None:
            if attachment_block.file_name == "user_audio_comment.ogg" and attachment_block.mime == r"audio/ogg":
                self.audio_comment = attachment_block.extract()
                field = QtWidgets.QTreeWidgetItem()
                field.setText(0, "Audio comment")
                widget = QtWidgets.QWidget()
                layout = QtWidgets.QHBoxLayout()
                widget.setLayout(layout)

                icon = QtGui.QIcon()
                icon.addPixmap(QtGui.QPixmap(":/play.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
                if icon.isNull():
                    play_audio_comment_btn = QtWidgets.QPushButton("Play comment")
                else:
                    play_audio_comment_btn = QtWidgets.QPushButton(widget)
                    play_audio_comment_btn.setIcon(icon)

                play_audio_comment_btn.clicked.connect(self.play_audio_comment)
                layout.addWidget(play_audio_comment_btn)

                self.audio_progress = QtWidgets.QProgressBar(widget)
                self.audio_progress.setValue(0)
                self.audio_progress.setTextVisible(False)
                layout.addWidget(self.audio_progress)

                layout.setStretch(1, 1)
                layout.setStretch(0, 0)

                self.fields.addTopLevelItem(field)
                self.fields.setItemWidget(field, 1, widget)

    def extract(self, event=None):
        attachment = self.file.mdf.attachments[self.index]
        encryption_info = extract_encryption_information(attachment.comment)
        password = None
        if encryption_info.get("encrypted", False) and self.file.mdf._mdf._password is None:
            text, ok = QtWidgets.QInputDialog.getText(
                self,
                "Attachment password",
                "The attachment is encrypted. Please provide the password:",
                QtWidgets.QLineEdit.EchoMode.Password,
            )
            if ok and text:
                password = text

        data, file_path, md5_sum = self.file.mdf.extract_attachment(self.index, password=password)

        file_name, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Select extracted file",
            str(file_path),
            "All files (*.*)",
            "All files (*.*)",
        )
        if file_name:
            file_name = Path(file_name)
            file_name.write_bytes(data)

    def play_audio_comment(self):
        if self.audio_comment is None:
            return

        global current_frame

        tmp = Path(gettempdir()) / "daxil_audio_comment.ogg"
        tmp.write_bytes(self.audio_comment)

        data, fs = sf.read(tmp)
        size = len(data)

        current_frame = 0
        playback_finished = threading.Event()

        def callback(outdata, frames, time, status):
            global current_frame
            chunksize = min(len(data) - current_frame, frames)
            outdata[:chunksize] = data[current_frame : current_frame + chunksize]
            if chunksize < frames:
                outdata[chunksize:] = 0
                raise sd.CallbackStop()
            current_frame += chunksize
            self.audio_progress.setValue(int(100 * current_frame / size))

        stream = sd.OutputStream(
            samplerate=fs, channels=data.shape[1], callback=callback, finished_callback=playback_finished.set
        )
        with stream:
            playback_finished.wait()

        self.audio_progress.setValue(0)

        tmp.unlink()



#!/usr/bin/env python
from pathlib import Path
import tempfile
import unittest

import numpy as np

from asammdf import MDF, Signal
from asammdf.blocks.mdf_v2 import MDF2
from asammdf.blocks.mdf_v3 import MDF3

CHANNEL_LEN = 10000


class TestMDF23(unittest.TestCase):
    tempdir: tempfile.TemporaryDirectory[str]

    @classmethod
    def setUpClass(cls) -> None:
        cls.tempdir = tempfile.TemporaryDirectory()

    def test_measurement(self) -> None:
        self.assertTrue(MDF2)
        self.assertTrue(MDF3)

    def test_read_mdf2_00(self) -> None:
        seed = np.random.randint(0, 2**31)

        np.random.seed(seed)
        print("Read 2.00 using seed =", seed)

        sig_int = Signal(
            np.random.randint(-(2**15), -1, CHANNEL_LEN),
            np.arange(CHANNEL_LEN),
            name="Integer Channel",
            unit="unit1",
        )

        sig_float = Signal(
            np.random.random(CHANNEL_LEN),
            np.arange(CHANNEL_LEN),
            name="Float Channel",
            unit="unit2",
        )

        with MDF(version="2.00") as mdf:
            mdf.append([sig_int, sig_float], common_timebase=True)
            outfile = mdf.save(Path(TestMDF23.tempdir.name) / "tmp", overwrite=True)

        with MDF(outfile) as mdf:
            ret_sig_int = mdf.get(sig_int.name)
            ret_sig_float = mdf.get(sig_float.name)

        self.assertTrue(np.array_equal(ret_sig_int.samples, sig_int.samples))
        self.assertTrue(np.array_equal(ret_sig_float.samples, sig_float.samples))

    def test_read_mdf2_14(self) -> None:
        seed = np.random.randint(0, 2**31)

        np.random.seed(seed)
        print("Read 2.14 using seed =", seed)

        sig_int = Signal(
            np.random.randint(-(2**29), 2**29, CHANNEL_LEN),
            np.arange(CHANNEL_LEN),
            name="Integer Channel",
            unit="unit1",
        )

        sig_float = Signal(
            np.random.random(CHANNEL_LEN),
            np.arange(CHANNEL_LEN),
            name="Float Channel",
            unit="unit2",
        )

        with MDF(version="2.14") as mdf:
            mdf.append([sig_int, sig_float], common_timebase=True)
            outfile = mdf.save(Path(TestMDF23.tempdir.name) / "tmp", overwrite=True)

        with MDF(outfile) as mdf:
            ret_sig_int = mdf.get(sig_int.name)
            ret_sig_float = mdf.get(sig_float.name)

        self.assertTrue(np.array_equal(ret_sig_int.samples, sig_int.samples))
        self.assertTrue(np.array_equal(ret_sig_float.samples, sig_float.samples))

    def test_read_mdf3_00(self) -> None:
        seed = np.random.randint(0, 2**31)

        np.random.seed(seed)
        print("Read 3.00 using seed =", seed)

        sig_int = Signal(
            np.random.randint(-(2**16), 2**16, CHANNEL_LEN, np.int32),
            np.arange(CHANNEL_LEN),
            name="Integer Channel",
            unit="unit1",
        )

        sig_float = Signal(
            np.random.random(CHANNEL_LEN),
            np.arange(CHANNEL_LEN),
            name="Float Channel",
            unit="unit2",
        )

        with MDF(version="3.00") as mdf:
            mdf.append([sig_int, sig_float], common_timebase=True)
            outfile = mdf.save(Path(TestMDF23.tempdir.name) / "tmp", overwrite=True)

        with MDF(outfile) as mdf:
            ret_sig_int = mdf.get(sig_int.name)
            ret_sig_float = mdf.get(sig_float.name)

        self.assertTrue(np.array_equal(ret_sig_int.samples, sig_int.samples))
        self.assertTrue(np.array_equal(ret_sig_float.samples, sig_float.samples))

    def test_read_mdf3_10(self) -> None:
        seed = np.random.randint(0, 2**31)

        np.random.seed(seed)
        print("Read 3.10 using seed =", seed)

        sig_int = Signal(
            np.random.randint(-(2**9), 2**7, CHANNEL_LEN),
            np.arange(CHANNEL_LEN),
            name="Integer Channel",
            unit="unit1",
        )

        sig_float = Signal(
            np.random.random(CHANNEL_LEN),
            np.arange(CHANNEL_LEN),
            name="Float Channel",
            unit="unit2",
        )

        with MDF(version="3.10") as mdf:
            mdf.append([sig_int, sig_float], common_timebase=True)
            outfile = mdf.save(Path(TestMDF23.tempdir.name) / "tmp", overwrite=True)

        with MDF(outfile) as mdf:
            ret_sig_int = mdf.get(sig_int.name)
            ret_sig_float = mdf.get(sig_float.name)

        self.assertTrue(np.array_equal(ret_sig_int.samples, sig_int.samples))
        self.assertTrue(np.array_equal(ret_sig_float.samples, sig_float.samples))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
import codecs
from pathlib import Path
import sys
from unittest import mock
import urllib
import urllib.request
from zipfile import ZipFile

import numpy as np
import pandas as pd
from PySide6 import QtCore, QtTest

from test.asammdf.gui.test_base import DBC, OpenMDF
from test.asammdf.gui.widgets.test_BaseBatchWidget import TestBatchWidget

# Note: If it's possible and make sense, use self.subTests
# to avoid initializing widgets multiple times and consume time.


class TestPushButtons(TestBatchWidget):
    def setUp(self):
        super().setUp()
        url = "https://github.com/danielhrisca/asammdf/files/4328945/OBD2-DBC-MDF4.zip"
        urllib.request.urlretrieve(url, "test.zip")
        ZipFile(r"test.zip").extractall(self.test_workspace)
        Path("test.zip").unlink()

        temp_dir = Path(self.test_workspace)

        # Get test files path
        self.mdf_path = [input_file for input_file in temp_dir.iterdir() if input_file.suffix == ".mf4"][0]
        self.dbc_path = [input_file for input_file in temp_dir.iterdir() if input_file.suffix == ".dbc"][0]

        self.setUpBatchWidget(measurement_files=[str(self.mdf_path)])
        # Go to Tab: "Bus logging": Index 3
        self.widget.aspects.setCurrentIndex(self.bus_aspect)

        # Ensure that CAN & LIN lists are empty
        self.assertEqual(self.widget.can_database_list.count(), 0)
        self.assertEqual(self.widget.lin_database_list.count(), 0)

    def load_database(self, path: Path | str | None = None, is_can=True):
        """
        Load database to CAN/LIN database list
        Parameters
        ----------
        path := path to .dbc file; if None: default .dbc path will be used
        is_can := CAN or LIN
        """
        if not path:
            path = self.dbc_path
        with mock.patch("asammdf.gui.widgets.batch.QtWidgets.QFileDialog.getOpenFileNames") as mo_getOpenFileNames:
            mo_getOpenFileNames.return_value = [str(path)], None
            if is_can:
                QtTest.QTest.mouseClick(self.widget.load_can_database_btn, QtCore.Qt.MouseButton.LeftButton)
            else:
                QtTest.QTest.mouseClick(self.widget.load_lin_database_btn, QtCore.Qt.MouseButton.LeftButton)

    def get_prefix(self):
        """
        Get prefix for channels names.
        Example:
            - 'CAN1.OBD2.'
        Returns
        -------
        prefix: str
        """
        # Prepare expected results
        # Linux cannot decode using 'utf-8' standard codec
        with OpenMDF(self.mdf_path) as mdf_file, codecs.open(str(self.dbc_path), encoding="ISO-8859-1") as dbc_file:
            for key, value in mdf_file.bus_logging_map.items():
                if value:
                    prefix = key
                    for new_key, new_value in value.items():
                        if value:
                            prefix += str(new_key)
            self.assertTrue(prefix)  # there is a suffix for feature channels
            prefix += "." + DBC.BO(dbc_file.readlines()).name + "."
            return prefix

    def test_load_can_database_btn(self):
        """
        Events:
            - Press Load CAN database button.

        Evaluate:
            - There is one item in can database list
            - The item's text is .dbc path
        """
        # Event
        self.load_database()
        # Evaluate
        self.assertEqual(self.widget.can_database_list.count(), 1)
        self.assertEqual(
            self.widget.can_database_list.itemWidget(self.widget.can_database_list.item(0)).database.text(),
            str(self.dbc_path),
        )

    def test_load_lin_database_btn(self):
        """
        Events:
            - Press Load LIN database button.

        Evaluate:
            - There is one item in lin database list
            - The item's text is .dbc path
        """
        # Event
        self.load_database(is_can=False)
        # Evaluate
        self.assertEqual(self.widget.lin_database_list.count(), 1)
        self.assertEqual(
            self.widget.lin_database_list.itemWidget(self.widget.lin_database_list.item(0)).database.text(),
            str(self.dbc_path),
        )

    def test_extract_bus_btn(self):
        """
        When QThreads are running, event-loops needs to be processed.
        Events:
            - Set prefix text
            - Press PushButton Extract Bus signals.

        Evaluate:
            - File was created.
            - New channels was created from .dbc and .mf4 files.
            - There is no channel from original measurement file, only from .dbc file
        """
        # Expected result
        output_file = Path.with_suffix(self.mdf_path, f".bus_logging{self.mdf_path.suffix}")

        # Precondition
        self.load_database()
        self.assertFalse(output_file.exists())
        self.assertEqual(self.widget.output_info_bus.toPlainText(), "")  # bus output info tab is clean

        # Set Prefix
        prefix = self.id().split(".")[-1]
        self.widget.prefix.setText(prefix)

        # Event
        self.mouse_click_on_btn_with_progress(self.widget.extract_bus_btn)

        self.assertTrue(output_file.exists())
        # Evaluate bus output info
        self.assertIn(str(self.dbc_path), self.widget.output_info_bus.toPlainText())
        self.assertIn(str(self.mdf_path), self.widget.output_info_bus.toPlainText())

        # because linux UnicodeDecodeError: 'utf-8' codec can't decode byte 0xb0 in position 943: invalid start byte
        with (
            OpenMDF(output_file) as bus_file,
            OpenMDF(self.mdf_path) as mdf_file,
            codecs.open(str(self.dbc_path), encoding="ISO-8859-1") as dbc_file,
        ):
            dbc_lines = dbc_file.readlines()
            source = DBC.BO(dbc_lines)

            self.assertNotEqual(len(mdf_file.groups), len(bus_file.groups))
            for channel in bus_file.channels_db:
                self.assertNotIn(channel.replace(prefix, ""), mdf_file.channels_db)

            # Evaluate Source:
            for name, group_index in mdf_file.channels_db.items():
                group, index = group_index[0]
                if name.endswith("ID"):
                    self.assertEqual(mdf_file.get(name, group, index).samples[0], int(source.id))
                if name.endswith("DataLength"):
                    self.assertEqual(mdf_file.get(name, group, index).samples[0], int(source.data_length))
                if "time" in name.lower():
                    max_from_mdf = mdf_file.get(name, group, index).samples[-1]
                    min_from_mdf = mdf_file.get(name, group, index).samples[0]
                    size_from_mdf = mdf_file.get(name, group, index).samples.size

            timestamps_size = set()
            timestamps_min = set()
            timestamps_max = set()
            for name, group_index in bus_file.channels_db.items():
                group, index = group_index[0]
                if not name.endswith("time"):
                    self.assertIn(prefix, name)

                    channel = bus_file.get(name, group, index, raw=True)
                    signal = DBC.SG(name.replace(self.id().split(".")[-1], "").split(".")[-1], dbc_lines)

                    if not ("service" in name or "response" in name):
                        self.assertEqual(channel.bit_count, signal.bit_count, name)

                    self.assertEqual(channel.unit, signal.unit)

                    if hasattr(channel.conversion, "a"):
                        self.assertEqual(channel.conversion.a, signal.conversion_a, name)
                        self.assertEqual(channel.conversion.b, signal.conversion_b, name)

                    timestamps_size.add(channel.timestamps.size)
                    timestamps_min.add(channel.timestamps.min())
                    timestamps_max.add(channel.timestamps.max())

                for group in bus_file.groups:
                    self.assertEqual(self.id().split(".")[-1], group.channel_group.acq_name.split(":")[0])

                for group in mdf_file.groups:
                    self.assertNotIn(self.id().split(".")[-1], group.channel_group.acq_name)

            self.assertEqual(max(timestamps_size), size_from_mdf)
            self.assertEqual(min(timestamps_min), min_from_mdf)
            self.assertEqual(max(timestamps_max), max_from_mdf)

    def test_extract_bus_csv_btn_0(self):
        """
        When QThreads are running, event-loops needs to be processed.
        This test will use mf4 generated file for feature evaluation.
        If `test_extract_bus_btn` is failed: this test must be ignored.

        Events:
            - Set prefix text
            - Ensure that all checkboxes are unchecked
            - Press PushButton Export to csv.

        Evaluate:
            - CSV files was created.
            - New channels was created from .dbc and .mf4 files
            - CSV files is mf4 channels groups
            - Channels names and groups names contain prefix
            - Signals samples are added to CSV
        """
        # Prepare expected results
        prefix = self.get_prefix()
        to_replace = [" ", '"', ":"]

        # Precondition
        self.load_database()
        for file in Path(self.test_workspace).iterdir():
            self.assertNotEqual(file.suffix, ".csv")
        self.assertEqual(self.widget.output_info_bus.toPlainText(), "")
        self.toggle_checkboxes(widget=self.widget.extract_bus_tab, check=False)  # uncheck all checkboxes

        # Expected results
        compare_to_file_path = Path.with_suffix(self.mdf_path, ".bus_logging.mf4")

        # Set Prefix
        self.widget.prefix.setText(self.id().split(".")[-1])

        # Get new mdf file
        self.mouse_click_on_btn_with_progress(self.widget.extract_bus_btn)

        # Event
        self.mouse_click_on_btn_with_progress(self.widget.extract_bus_csv_btn)

        csv_tables = [
            (file, pd.read_csv(file)) for file in sorted(Path(self.test_workspace).iterdir()) if file.suffix == ".csv"
        ]
        with OpenMDF(compare_to_file_path) as mdf_file:
            for group, (path, table) in zip(mdf_file.groups, csv_tables, strict=False):
                comment = group.channel_group.comment
                for char in to_replace:
                    comment = comment.replace(char, "_")

                # Evaluate CSV name
                self.assertIn(self.id().split(".")[-1], str(path))
                self.assertTrue(str(path).endswith(comment + ".csv"))
                self.assertTrue(str(path.stem).startswith(compare_to_file_path.stem))

                # Evaluate channels
                for channel in group.channels:
                    if channel.name != "time":
                        name = prefix + channel.name

                        self.assertIn(name, table.columns)
                        ch = mdf_file.get(channel.name)

                        # Evaluate samples
                        self.assertEqual(ch.samples.size, table[name].size)
                        if np.issubdtype(ch.samples.dtype, np.number):
                            self.assertAlmostEqual(ch.samples.min(), table[name].min(), places=10)
                            self.assertAlmostEqual(ch.samples.max(), table[name].max(), places=10)
                        else:
                            self.assertEqual(
                                str(min(ch.samples)).replace(" ", ""),
                                str(table[name].min()).replace(" ", ""),
                            )
                            self.assertEqual(
                                str(max(ch.samples)).replace(" ", ""),
                                str(table[name].max()).replace(" ", ""),
                            )

                        # Evaluate timestamps
                        self.assertEqual(ch.timestamps.size, table.timestamps.size)
                        self.assertAlmostEqual(ch.timestamps.min(), table.timestamps.min(), places=10)
                        self.assertAlmostEqual(ch.timestamps.max(), table.timestamps.max(), places=10)

    def test_extract_bus_csv_btn_1(self):
        """
        When QThreads are running, event-loops needs to be processed.
        This test will use mf4 generated file for feature evaluation.
        If `test_extract_bus_btn` is failed: this test must be ignored.

        Events:
            - Set prefix text
            - Ensure that all checkboxes are checked
            - Press PushButton Export to csv.

        Evaluate:
            - CSV file is created
            - Timestamps start from 0 + measurement start time
            - CSV file contain new channels from .mf4 file
            - Channels names contain prefix
            - Signals samples are added to CSV
        """
        prefix = self.get_prefix()

        # Precondition
        self.load_database()
        for file in Path(self.test_workspace).iterdir():
            self.assertNotEqual(file.suffix, ".csv")
        self.assertEqual(self.widget.output_info_bus.toPlainText(), "")
        self.toggle_checkboxes(widget=self.widget.extract_bus_tab, check=True)  # uncheck all checkboxes

        # Expected results
        csv_path = Path.with_suffix(self.mdf_path, ".bus_logging.csv")
        compare_to_file_path = Path.with_suffix(self.mdf_path, ".bus_logging.mf4")

        # Set Prefix
        self.widget.prefix.setText(self.id().split(".")[-1])

        # Get new mdf file
        self.mouse_click_on_btn_with_progress(self.widget.extract_bus_btn)

        # Event
        self.mouse_click_on_btn_with_progress(self.widget.extract_bus_csv_btn)

        csv_table = pd.read_csv(csv_path, header=[0, 1], engine="python", encoding="ISO-8859-1")
        with OpenMDF(compare_to_file_path) as mdf_file:
            # Get channels timestamps max, min, difference between extremes
            min_ts = min(channel.timestamps.min() for channel in mdf_file.iter_channels())
            max_ts = max(channel.timestamps.max() for channel in mdf_file.iter_channels())
            seconds = int(max_ts - min_ts)
            microseconds = np.floor((max_ts - min_ts - seconds) * pow(10, 6))

            delay = np.datetime64(mdf_file.start_time)  # - np.timedelta64(3, "h")   # idk from why, local fail...
            csv_table.timestamps = pd.DatetimeIndex(csv_table.timestamps.values, yearfirst=True).values - delay

            # Evaluate timestamps min
            self.assertEqual(np.timedelta64(csv_table.timestamps.values.min(), "us").item().microseconds, 0)
            self.assertEqual(np.timedelta64(csv_table.timestamps.values.min(), "us").item().seconds, 0)
            # Evaluate timestamps max
            self.assertEqual(
                np.timedelta64(csv_table.timestamps.values.max(), "us").item().microseconds, int(microseconds)
            )
            self.assertEqual(np.timedelta64(csv_table.timestamps.values.max(), "us").item().seconds, seconds)

            for ch in mdf_file.iter_channels():
                name = prefix + ch.name
                self.assertIn(name, csv_table.columns)
                column = csv_table[name]
                if ch.unit and (sys.platform == "win32"):
                    self.assertEqual(ch.unit, column.columns[0])

                # Evaluate channel samples
                if np.issubdtype(ch.samples.dtype, np.number):
                    self.assertAlmostEqual(ch.samples.min(), column.values.min(), places=10)
                    self.assertAlmostEqual(ch.samples.max(), column.values.max(), places=10)
                else:
                    self.assertEqual(
                        str(min(ch.samples)).replace(" ", ""),
                        str(column.values.min()).replace(" ", ""),
                    )
                    self.assertEqual(
                        str(max(ch.samples)).replace(" ", ""),
                        str(column.values.max()).replace(" ", ""),
                    )



#!/usr/bin/env python
from pathlib import Path
from unittest import mock

import numpy as np

from test.asammdf.gui.test_base import OpenMDF
from test.asammdf.gui.widgets.test_BaseBatchWidget import TestBatchWidget

# Note: If it's possible and make sense, use self.subTests
# to avoid initializing widgets multiple times and consume time.


class TestPushButtons(TestBatchWidget):
    test_file_0_name = "test_batch_cut_0.mf4"
    test_file_1_name = "test_batch_cut_1.mf4"
    output_file_name = "output.mf4"

    def setUp(self):
        super().setUp()

        self.test_file_0 = Path(self.test_workspace, self.test_file_0_name)
        self.test_file_1 = Path(self.test_workspace, self.test_file_1_name)

        self.copy_mdf_files_to_workspace()
        self.setUpBatchWidget(measurement_files=[str(self.test_file_0), str(self.test_file_1)])

        # Go to Tab: Concatenate
        self.widget.aspects.setCurrentIndex(self.concatenate_aspect)
        self.processEvents(0.1)

    def test_PushButton_Concatenate(self):
        """
        Events:
            - Open 'BatchWidget' with created measurement.
            - Go to Tab: "Concatenate": Index 0
            - Press PushButton "Concatenate"

        Evaluate:
            - New file is created
            - All channels samples and timestamps from test files are concatenated intro selected output file
        """
        # Get evaluation data
        output_file = Path(self.test_workspace, self.output_file_name)

        # Event
        with mock.patch("asammdf.gui.widgets.batch.QtWidgets.QFileDialog.getSaveFileName") as mo_getSaveFileName:
            mo_getSaveFileName.return_value = output_file, ""
            self.mouse_click_on_btn_with_progress(self.widget.concatenate_btn)

        # Evaluate that file exist
        self.assertTrue(output_file.exists())

        # Evaluate
        with (
            OpenMDF(output_file) as new_mdf_file,
            OpenMDF(self.test_file_0) as original_mdf_0,
            OpenMDF(self.test_file_1) as original_mdf_1,
        ):
            # Evaluate saved file
            for ch in new_mdf_file.iter_channels():
                if ch.name in original_mdf_0.channels_db:  # avoid `__samples_origin` channel
                    ch_0 = original_mdf_0.get(ch.name)
                    ch_1 = original_mdf_1.get(ch.name)

                    self.assertTrue(np.array_equal(np.concatenate([ch_0.samples, ch_1.samples]), ch.samples))
                    self.assertTrue(np.array_equal(np.concatenate([ch_0.timestamps, ch_1.timestamps]), ch.timestamps))



"""asammdf `Signal` class module for time-correct signal processing"""

from collections.abc import Iterator
import logging
from pathlib import Path
from textwrap import fill
from typing import TYPE_CHECKING, Union

import numpy as np
from numpy.typing import ArrayLike, DTypeLike, NDArray
from typing_extensions import Any, overload

from .blocks import v2_v3_blocks as v3b
from .blocks import v4_blocks as v4b
from .blocks.conversion_utils import from_dict
from .blocks.options import FloatInterpolation, IntegerInterpolation
from .blocks.source_utils import Source
from .blocks.types import (
    ChannelConversionType,
    FloatInterpolationModeType,
    IntInterpolationModeType,
    SourceType,
)
from .blocks.utils import extract_xml_comment, MdfException, SignalFlags
from .version import __version__

if TYPE_CHECKING:
    from mpl_toolkits.mplot3d.art3d import Line3DCollection

try:
    encode = np.strings.encode
except:
    encode = np.char.encode

logger = logging.getLogger("asammdf")

ORIGIN_UNKNOWN = (-1, -1)


class InvalidationArray(np.ndarray[tuple[int], np.dtype[np.bool]]):
    ORIGIN_UNKNOWN = ORIGIN_UNKNOWN

    def __new__(cls, input_array: ArrayLike, origin: tuple[int, int] = ORIGIN_UNKNOWN) -> "InvalidationArray":
        obj = np.asarray(input_array).view(cls)
        obj.origin = origin
        return obj

    def __array_finalize__(self, obj: NDArray[np.bool] | None) -> None:
        if obj is None:
            return
        self.origin: tuple[int, int] = getattr(obj, "origin", ORIGIN_UNKNOWN)


class Signal:
    """The `Signal` represents a channel described by its samples and
    timestamps. It can perform arithmetic operations against other `Signal`
    objects or numeric types. The operations are computed in respect to the
    timestamps (time-correct). The non-float signals are not interpolated,
    instead the last value relative to the current timestamp is used.
    `samples`, `timestamps` and `name` are mandatory arguments.

    Parameters
    ----------
    samples : array-like
        Signal samples.
    timestamps : array-like
        Signal timestamps.
    unit : str, optional
        Signal unit.
    name : str
        Signal name.
    conversion : dict | channel conversion block, optional
        Dict that contains extra conversion information about the signal.
    comment : str, optional
        Signal comment.
    raw : bool, default True
        Signal samples are raw values, with no physical conversion applied.
    master_metadata : tuple, optional
        Master name and sync type.
    display_names : dict, optional
        Display names used by MDF version 3.
    attachment : tuple, optional
        Channel attachment and name from MDF version 4.
    source : Source, optional
        Source information named tuple.
    bit_count : int, optional
        Bit count; useful for integer channels.
    invalidation_bits : array-like, optional
        Channel invalidation bits.
    encoding : str, optional
        Encoding for string signals.
    flags : int, optional
        Flags for user-defined attributes and stream sync.
    """

    Flags = SignalFlags

    def __init__(
        self,
        samples: ArrayLike,
        timestamps: ArrayLike,
        unit: str = "",
        name: str = "",
        conversion: dict[str, object] | ChannelConversionType | None = None,
        comment: str = "",
        raw: bool = True,
        master_metadata: tuple[str, int] | None = None,
        display_names: dict[str, str] | None = None,
        attachment: tuple[bytes | str, Path, bytes | str] | None = None,
        source: SourceType | None = None,
        bit_count: int | None = None,
        invalidation_bits: ArrayLike | None = None,
        encoding: str | None = None,
        group_index: int = -1,
        channel_index: int = -1,
        flags: int = Flags.no_flags,
        virtual_conversion: dict[str, object] | ChannelConversionType | None = None,
        virtual_master_conversion: dict[str, object] | ChannelConversionType | None = None,
    ) -> None:
        if not name:
            message = (
                '"samples", "timestamps" and "name" are mandatory '
                f"for Signal class __init__: samples={samples!r}\n"
                f"timestamps={timestamps!r}\nname={name}"
            )
            raise MdfException(message)
        else:
            self.samples: NDArray[Any]
            if not isinstance(samples, np.ndarray):
                samples = np.array(samples)
                kind = samples.dtype.kind
                if kind == "U":
                    if encoding is None:
                        encodings = ["utf-8", "latin-1"]
                    else:
                        encodings = [encoding, "utf-8", "latin-1"]
                    for _encoding in encodings:
                        try:
                            self.samples = encode(samples, _encoding)
                            break
                        except:
                            continue
                    else:
                        self.samples = encode(samples, encodings[0], errors="ignore")
                elif kind == "O":
                    self.samples = samples.astype(np.bytes_)
                else:
                    self.samples = samples
            else:
                self.samples = samples

            self.timestamps: NDArray[Any]
            if not isinstance(timestamps, np.ndarray):
                self.timestamps = np.array(timestamps, dtype=np.float64)
            else:
                self.timestamps = timestamps

            if self.samples.shape[0] != self.timestamps.shape[0]:
                message = "{} samples and timestamps length mismatch ({} vs {})"
                message = message.format(name, self.samples.shape[0], self.timestamps.shape[0])
                logger.exception(message)
                raise MdfException(message)

            self.unit = unit
            self.name = name
            self.comment = comment
            self.flags = flags
            self._plot_axis: Line3DCollection | None = None
            self.raw = raw
            self.master_metadata = master_metadata
            self.display_names = display_names or {}
            self.attachment = attachment
            self.encoding = encoding
            self.group_index = group_index
            self.channel_index = channel_index
            self._invalidation_bits = InvalidationArray(invalidation_bits) if invalidation_bits is not None else None

            self.source: Source | None
            if source:
                if not isinstance(source, Source):
                    self.source = Source.from_source(source)
                else:
                    self.source = source
            else:
                self.source = None

            if bit_count is None:
                self.bit_count = self.samples.dtype.itemsize * 8
            else:
                self.bit_count = bit_count

            self.conversion: ChannelConversionType | None
            if conversion:
                if not isinstance(conversion, (v4b.ChannelConversion, v3b.ChannelConversion)):
                    self.conversion = from_dict(conversion)
                else:
                    self.conversion = conversion
            else:
                self.conversion = None

            self.virtual_conversion: ChannelConversionType | None
            if self.flags & self.Flags.virtual:
                if not isinstance(virtual_conversion, (v4b.ChannelConversion, v3b.ChannelConversion)):
                    self.virtual_conversion = from_dict(virtual_conversion)
                else:
                    self.virtual_conversion = virtual_conversion
            else:
                self.virtual_conversion = None

            self.virtual_master_conversion: ChannelConversionType | None
            if self.flags & self.Flags.virtual_master:
                if not isinstance(virtual_master_conversion, (v4b.ChannelConversion, v3b.ChannelConversion)):
                    self.virtual_master_conversion = from_dict(virtual_master_conversion)
                else:
                    self.virtual_master_conversion = virtual_master_conversion
            else:
                self.virtual_master_conversion = None

    @property
    def invalidation_bits(self) -> InvalidationArray | None:
        return self._invalidation_bits

    @invalidation_bits.setter
    def invalidation_bits(self, value: ArrayLike | None) -> None:
        if value is None:
            self._invalidation_bits = None

        else:
            if not isinstance(value, InvalidationArray):
                value = InvalidationArray(value)

            if value.shape[0] != self.samples.shape[0]:
                message = "{} samples and invalidation bits length mismatch ({} vs {})"
                message = message.format(self.name, self.samples.shape[0], value.shape[0])
                logger.exception(message)
                raise MdfException(message)

            self._invalidation_bits = value

    def __repr__(self) -> str:
        return f"""<Signal {self.name}:
\tsamples={self.samples}
\ttimestamps={self.timestamps}
\tinvalidation_bits={self.invalidation_bits}
\tunit="{self.unit}"
\tconversion={self.conversion}
\tsource={self.source}
\tcomment="{self.comment}"
\tflags="{self.flags}"
\tmastermeta="{self.master_metadata}"
\traw={self.raw}
\tdisplay_names={self.display_names}
\tattachment={self.attachment}>
"""

    def plot(self, validate: bool = True, index_only: bool = False) -> None:
        """Plot Signal samples. Pyqtgraph is used if it is available; in this
        case see the GUI plot documentation to see the available commands.

        Parameters
        ----------
        validate : bool, default True
            Apply the invalidation bits.
        index_only : bool, default False
            Use index-based X axis. This can be useful if the master (usually
            time based) is corrupted with NaN, inf or if it is not strictly
            increasing.
        """
        try:
            from .gui.plot import plot

            plot(self, validate=True, index_only=False)
            return

        except:
            try:
                import matplotlib.pyplot as plt
                from matplotlib.widgets import Slider
                from mpl_toolkits.mplot3d import Axes3D
            except ImportError:
                logging.warning("Signal plotting requires pyqtgraph or matplotlib")
                return

        if len(self.samples.shape) <= 1 and self.samples.dtype.names is None:
            fig = plt.figure()
            if fig.canvas.manager:
                fig.canvas.manager.set_window_title(self.name)
            fig.text(
                0.95,
                0.05,
                f"asammdf {__version__}",
                fontsize=8,
                color="red",
                ha="right",
                va="top",
                alpha=0.5,
            )

            name = self.name

            if self.comment:
                comment = self.comment.replace("$", "")
                comment = extract_xml_comment(comment)
                comment = fill(comment, 120).replace("\\n", " ")

                title = f"{name}\n({comment})"
                plt.title(title)
            else:
                plt.title(name)
            try:
                if not self.master_metadata:
                    plt.xlabel("Time [s]")
                    plt.ylabel(f"[{self.unit}]")
                    plt.plot(self.timestamps, self.samples, "b")
                    plt.plot(self.timestamps, self.samples, "b.")
                    plt.grid(True)
                    plt.show()
                else:
                    master_name, sync_type = self.master_metadata
                    match sync_type:
                        case 0 | 1:
                            plt.xlabel(f"{master_name} [s]")
                        case 2:
                            plt.xlabel(f"{master_name} [deg]")
                        case 3:
                            plt.xlabel(f"{master_name} [m]")
                        case 4:
                            plt.xlabel(f"{master_name} [index]")
                    plt.ylabel(f"[{self.unit}]")
                    plt.plot(self.timestamps, self.samples, "b")
                    plt.plot(self.timestamps, self.samples, "b.")
                    plt.grid(True)
                    plt.show()
            except ValueError:
                plt.close(fig)
        else:
            try:
                names = self.samples.dtype.names
                if names is None or len(names) == 1:
                    if names:
                        samples = self.samples[names[0]]
                    else:
                        samples = self.samples

                    shape = samples.shape[1:]

                    fig = plt.figure()
                    if fig.canvas.manager:
                        fig.canvas.manager.set_window_title(self.name)
                    fig.text(
                        0.95,
                        0.05,
                        f"asammdf {__version__}",
                        fontsize=8,
                        color="red",
                        ha="right",
                        va="top",
                        alpha=0.5,
                    )

                    if self.comment:
                        comment = self.comment.replace("$", "")
                        plt.title(f"{self.name}\n({comment})")
                    else:
                        plt.title(self.name)

                    ax: Axes3D = fig.add_subplot(111, projection="3d")

                    # Grab some test data.
                    X = np.array(range(shape[1]))
                    Y = np.array(range(shape[0]))
                    X, Y = np.meshgrid(X, Y)

                    Z = samples[0]

                    # Plot a basic wireframe.
                    self._plot_axis = plot_axis = ax.plot_wireframe(X, Y, Z, rstride=1, cstride=1)

                    # Place Sliders on Graph
                    ax_a = plt.axes((0.25, 0.1, 0.65, 0.03))

                    # Create Sliders & Determine Range
                    sa = Slider(
                        ax_a,
                        "Time [s]",
                        self.timestamps[0],
                        self.timestamps[-1],
                        valinit=self.timestamps[0],
                    )

                    def update(val: float) -> None:
                        plot_axis.remove()
                        idx = np.searchsorted(self.timestamps, sa.val, side="right")
                        Z = samples[idx - 1]
                        self._plot_axis = ax.plot_wireframe(X, Y, Z, rstride=1, cstride=1)
                        fig.canvas.draw_idle()

                    sa.on_changed(update)

                    plt.show()

                else:
                    fig = plt.figure()
                    if fig.canvas.manager:
                        fig.canvas.manager.set_window_title(self.name)
                    fig.text(
                        0.95,
                        0.05,
                        f"asammdf {__version__}",
                        fontsize=8,
                        color="red",
                        ha="right",
                        va="top",
                        alpha=0.5,
                    )

                    if self.comment:
                        comment = self.comment.replace("$", "")
                        plt.title(f"{self.name}\n({comment})")
                    else:
                        plt.title(self.name)

                    ax = fig.add_subplot(111, projection="3d")

                    samples = self.samples[names[0]]
                    axis1 = self.samples[names[1]]
                    axis2 = self.samples[names[2]]

                    # Grab some test data.
                    X, Y = np.meshgrid(axis2[0], axis1[0])

                    Z = samples[0]

                    # Plot a basic wireframe.
                    self._plot_axis = plot_axis = ax.plot_wireframe(X, Y, Z, rstride=1, cstride=1)

                    # Place Sliders on Graph
                    ax_a = plt.axes((0.25, 0.1, 0.65, 0.03))

                    # Create Sliders & Determine Range
                    sa = Slider(
                        ax_a,
                        "Time [s]",
                        self.timestamps[0],
                        self.timestamps[-1],
                        valinit=self.timestamps[0],
                    )

                    def update(val: float) -> None:
                        plot_axis.remove()
                        idx = np.searchsorted(self.timestamps, sa.val, side="right")
                        Z = samples[idx - 1]
                        X, Y = np.meshgrid(axis2[idx - 1], axis1[idx - 1])
                        self._plot_axis = ax.plot_wireframe(X, Y, Z, rstride=1, cstride=1)
                        fig.canvas.draw_idle()

                    sa.on_changed(update)

                    plt.show()

            except Exception as err:
                print(err)

    def cut(
        self,
        start: float | None = None,
        stop: float | None = None,
        include_ends: bool = True,
        integer_interpolation_mode: (
            IntInterpolationModeType | IntegerInterpolation
        ) = IntegerInterpolation.REPEAT_PREVIOUS_SAMPLE,
        float_interpolation_mode: (
            FloatInterpolationModeType | FloatInterpolation
        ) = FloatInterpolation.LINEAR_INTERPOLATION,
    ) -> "Signal":
        """Cut the signal according to the `start` and `stop` values, by using
        the insertion indexes in the signal's time axis.

        Parameters
        ----------
        start : float, optional
            Start timestamp for cutting.
        stop : float, optional
            Stop timestamp for cutting.
        include_ends : bool, default True
            Include the `start` and `stop` timestamps after cutting the signal.
            If `start` and `stop` are not found in the original timestamps,
            then the new samples will be computed using interpolation.
        integer_interpolation_mode : int, default 0
            Interpolation mode for integer signals.

            * 0 - repeat previous sample
            * 1 - linear interpolation
            * 2 - hybrid interpolation: channels with integer data type (raw
              values) that have a conversion that outputs float values will use
              linear interpolation, otherwise the previous sample is used

            .. versionadded:: 6.2.0

        float_interpolation_mode : int, default 1
            Interpolation mode for float channels.

            * 0 - repeat previous sample
            * 1 - linear interpolation

            .. versionadded:: 6.2.0

        Returns
        -------
        result : Signal
            New `Signal` cut from the original.

        Examples
        --------
        >>> from asammdf import Signal
        >>> import numpy as np
        >>> old_sig = Signal(np.arange(0.03, 100, 0.05), np.arange(0.03, 100, 0.05), name='SIG')
        >>> new_sig = old_sig.cut(1.0, 10.5)
        >>> new_sig.timestamps[0], new_sig.timestamps[-1]
        (1.0, 10.5)

        >>> new_sig = old_sig.cut(1.0, 10.5, include_ends=False)
        >>> new_sig.timestamps[0], new_sig.timestamps[-1]
        (1.03, 10.48)

        >>> new_sig = old_sig.cut(1.0, 10.5, float_interpolation_mode=0)
        >>> new_sig.samples[0], new_sig.samples[-1]
        (0.98, 10.48)
        """

        integer_interpolation_mode = IntegerInterpolation(integer_interpolation_mode)
        float_interpolation_mode = FloatInterpolation(float_interpolation_mode)

        if self.samples.size == 0:
            return Signal(
                np.array([], dtype=self.samples.dtype),
                np.array([], dtype=self.timestamps.dtype),
                self.unit,
                self.name,
                self.conversion,
                self.comment,
                self.raw,
                self.master_metadata,
                self.display_names,
                self.attachment,
                self.source,
                self.bit_count,
                encoding=self.encoding,
                group_index=self.group_index,
                channel_index=self.channel_index,
                flags=self.flags,
                virtual_conversion=self.virtual_conversion,
                virtual_master_conversion=self.virtual_master_conversion,
            )

        invalidation_bits: NDArray[np.bool] | None
        if start is None:
            if stop is None:
                # return the channel uncut
                result = Signal(
                    self.samples.copy(),
                    self.timestamps.copy(),
                    self.unit,
                    self.name,
                    self.conversion,
                    self.comment,
                    self.raw,
                    self.master_metadata,
                    self.display_names,
                    self.attachment,
                    self.source,
                    self.bit_count,
                    invalidation_bits=self.invalidation_bits.copy() if self.invalidation_bits is not None else None,
                    encoding=self.encoding,
                    group_index=self.group_index,
                    channel_index=self.channel_index,
                    flags=self.flags,
                    virtual_conversion=self.virtual_conversion,
                    virtual_master_conversion=self.virtual_master_conversion,
                )

            else:
                # cut from begining to stop
                if stop < self.timestamps[0]:
                    result = Signal(
                        np.array([], dtype=self.samples.dtype),
                        np.array([], dtype=self.timestamps.dtype),
                        self.unit,
                        self.name,
                        self.conversion,
                        self.comment,
                        self.raw,
                        self.master_metadata,
                        self.display_names,
                        self.attachment,
                        self.source,
                        self.bit_count,
                        encoding=self.encoding,
                        group_index=self.group_index,
                        channel_index=self.channel_index,
                        flags=self.flags,
                        virtual_conversion=self.virtual_conversion,
                        virtual_master_conversion=self.virtual_master_conversion,
                    )

                else:
                    stop_idx = np.searchsorted(self.timestamps, stop, side="right")
                    if include_ends and stop not in self.timestamps and stop < self.timestamps[-1]:
                        interpolated = self.interp(
                            [stop],
                            integer_interpolation_mode=integer_interpolation_mode,
                            float_interpolation_mode=float_interpolation_mode,
                        )

                        if len(interpolated):
                            samples = np.append(self.samples[:stop_idx], interpolated.samples, axis=0)
                            timestamps = np.append(self.timestamps[:stop_idx], interpolated.timestamps)
                            if self.invalidation_bits is not None and interpolated.invalidation_bits is not None:
                                invalidation_bits = InvalidationArray(
                                    np.append(
                                        self.invalidation_bits[:stop_idx],
                                        interpolated.invalidation_bits,
                                    ),
                                    self.invalidation_bits.origin,
                                )
                            else:
                                invalidation_bits = None
                    else:
                        samples = self.samples[:stop_idx].copy()
                        timestamps = self.timestamps[:stop_idx].copy()
                        if self.invalidation_bits is not None:
                            invalidation_bits = self.invalidation_bits[:stop_idx].copy()
                        else:
                            invalidation_bits = None

                    if samples.dtype != self.samples.dtype:
                        samples = samples.astype(self.samples.dtype)

                    result = Signal(
                        samples,
                        timestamps,
                        self.unit,
                        self.name,
                        self.conversion,
                        self.comment,
                        self.raw,
                        self.master_metadata,
                        self.display_names,
                        self.attachment,
                        self.source,
                        self.bit_count,
                        invalidation_bits=invalidation_bits,
                        encoding=self.encoding,
                        group_index=self.group_index,
                        channel_index=self.channel_index,
                        flags=self.flags,
                        virtual_conversion=self.virtual_conversion,
                        virtual_master_conversion=self.virtual_master_conversion,
                    )

        else:
            if stop is None:
                # cut from start to end
                if start > self.timestamps[-1]:
                    result = Signal(
                        np.array([], dtype=self.samples.dtype),
                        np.array([], dtype=self.timestamps.dtype),
                        self.unit,
                        self.name,
                        self.conversion,
                        self.comment,
                        self.raw,
                        self.master_metadata,
                        self.display_names,
                        self.attachment,
                        self.source,
                        self.bit_count,
                        encoding=self.encoding,
                        group_index=self.group_index,
                        channel_index=self.channel_index,
                        flags=self.flags,
                        virtual_conversion=self.virtual_conversion,
                        virtual_master_conversion=self.virtual_master_conversion,
                    )

                else:
                    start_idx = np.searchsorted(self.timestamps, start, side="left")
                    if include_ends and start not in self.timestamps and start > self.timestamps[0]:
                        interpolated = self.interp(
                            [start],
                            integer_interpolation_mode=integer_interpolation_mode,
                            float_interpolation_mode=float_interpolation_mode,
                        )
                        if len(interpolated):
                            samples = np.append(interpolated.samples, self.samples[start_idx:], axis=0)
                            timestamps = np.append(interpolated.timestamps, self.timestamps[start_idx:])
                            if self.invalidation_bits is not None and interpolated.invalidation_bits is not None:
                                invalidation_bits = InvalidationArray(
                                    np.append(
                                        interpolated.invalidation_bits,
                                        self.invalidation_bits[start_idx:],
                                    ),
                                    self.invalidation_bits.origin,
                                )
                            else:
                                invalidation_bits = None
                    else:
                        samples = self.samples[start_idx:].copy()
                        timestamps = self.timestamps[start_idx:].copy()
                        if self.invalidation_bits is not None:
                            invalidation_bits = self.invalidation_bits[start_idx:].copy()
                        else:
                            invalidation_bits = None

                    if samples.dtype != self.samples.dtype:
                        samples = samples.astype(self.samples.dtype)

                    result = Signal(
                        samples,
                        timestamps,
                        self.unit,
                        self.name,
                        self.conversion,
                        self.comment,
                        self.raw,
                        self.master_metadata,
                        self.display_names,
                        self.attachment,
                        self.source,
                        self.bit_count,
                        invalidation_bits=invalidation_bits,
                        encoding=self.encoding,
                        group_index=self.group_index,
                        channel_index=self.channel_index,
                        flags=self.flags,
                        virtual_conversion=self.virtual_conversion,
                        virtual_master_conversion=self.virtual_master_conversion,
                    )

            else:
                # cut between start and stop
                if start > self.timestamps[-1] or stop < self.timestamps[0]:
                    result = Signal(
                        np.array([], dtype=self.samples.dtype),
                        np.array([], dtype=self.timestamps.dtype),
                        self.unit,
                        self.name,
                        self.conversion,
                        self.comment,
                        self.raw,
                        self.master_metadata,
                        self.display_names,
                        self.attachment,
                        self.source,
                        self.bit_count,
                        encoding=self.encoding,
                        group_index=self.group_index,
                        channel_index=self.channel_index,
                        flags=self.flags,
                        virtual_conversion=self.virtual_conversion,
                        virtual_master_conversion=self.virtual_master_conversion,
                    )
                else:
                    start_idx = np.searchsorted(self.timestamps, start, side="left")
                    stop_idx = np.searchsorted(self.timestamps, stop, side="right")

                    if start_idx == stop_idx:
                        if include_ends:
                            if start == stop:
                                ends = np.array([start], dtype=self.timestamps.dtype)
                            else:
                                ends = np.array(
                                    [start, stop],
                                    dtype=self.timestamps.dtype,
                                )

                            interpolated = self.interp(
                                ends,
                                integer_interpolation_mode=integer_interpolation_mode,
                                float_interpolation_mode=float_interpolation_mode,
                            )
                            samples = interpolated.samples
                            timestamps = ends
                            invalidation_bits = interpolated.invalidation_bits
                        else:
                            samples = np.array([], dtype=self.samples.dtype)
                            timestamps = np.array([], dtype=self.timestamps.dtype)
                            if self.invalidation_bits is not None:
                                invalidation_bits = self.invalidation_bits[0:0]
                            else:
                                invalidation_bits = None
                    else:
                        samples = self.samples[start_idx:stop_idx].copy()
                        timestamps = self.timestamps[start_idx:stop_idx].copy()
                        if self.invalidation_bits is not None:
                            invalidation_bits = self.invalidation_bits[start_idx:stop_idx].copy()
                        else:
                            invalidation_bits = None

                        if include_ends and stop not in self.timestamps and stop < self.timestamps[-1]:
                            interpolated = self.interp(
                                [stop],
                                integer_interpolation_mode=integer_interpolation_mode,
                                float_interpolation_mode=float_interpolation_mode,
                            )

                            if len(interpolated):
                                samples = np.append(samples, interpolated.samples, axis=0)
                                timestamps = np.append(timestamps, interpolated.timestamps)
                                if invalidation_bits is not None and interpolated.invalidation_bits is not None:
                                    invalidation_bits = InvalidationArray(
                                        np.append(
                                            invalidation_bits,
                                            interpolated.invalidation_bits,
                                        ),
                                        interpolated.invalidation_bits.origin,
                                    )

                        if include_ends and start not in self.timestamps and start > self.timestamps[0]:
                            interpolated = self.interp(
                                [start],
                                integer_interpolation_mode=integer_interpolation_mode,
                                float_interpolation_mode=float_interpolation_mode,
                            )

                            if len(interpolated):
                                samples = np.append(interpolated.samples, samples, axis=0)
                                timestamps = np.append(interpolated.timestamps, timestamps)

                                if invalidation_bits is not None and interpolated.invalidation_bits is not None:
                                    invalidation_bits = InvalidationArray(
                                        np.append(
                                            interpolated.invalidation_bits,
                                            invalidation_bits,
                                        ),
                                        interpolated.invalidation_bits.origin,
                                    )

                    if samples.dtype != self.samples.dtype:
                        samples = samples.astype(self.samples.dtype)

                    result = Signal(
                        samples,
                        timestamps,
                        self.unit,
                        self.name,
                        self.conversion,
                        self.comment,
                        self.raw,
                        self.master_metadata,
                        self.display_names,
                        self.attachment,
                        self.source,
                        self.bit_count,
                        invalidation_bits=invalidation_bits,
                        encoding=self.encoding,
                        group_index=self.group_index,
                        channel_index=self.channel_index,
                        flags=self.flags,
                        virtual_conversion=self.virtual_conversion,
                        virtual_master_conversion=self.virtual_master_conversion,
                    )

        return result

    def extend(self, other: "Signal") -> "Signal":
        """Extend Signal with samples from another Signal.

        Parameters
        ----------
        other : Signal

        Returns
        -------
        signal : Signal
            New extended `Signal`.
        """
        if len(self.timestamps):
            last_stamp = self.timestamps[-1]
        else:
            last_stamp = 0
        if len(other):
            other_first_sample = other.timestamps[0]
            if last_stamp >= other_first_sample:
                timestamps = other.timestamps + last_stamp
            else:
                timestamps = other.timestamps

            if self.invalidation_bits is None:
                if other.invalidation_bits is None:
                    invalidation_bits = None
                else:
                    invalidation_bits = InvalidationArray(
                        np.concatenate((np.zeros(len(self), dtype=bool), other.invalidation_bits)),
                        other.invalidation_bits.origin,
                    )
            else:
                if other.invalidation_bits is None:
                    invalidation_bits = InvalidationArray(
                        np.concatenate((self.invalidation_bits, np.zeros(len(other), dtype=bool))),
                        self.invalidation_bits.origin,
                    )
                else:
                    invalidation_bits = InvalidationArray(
                        np.append(self.invalidation_bits, other.invalidation_bits), self.invalidation_bits.origin
                    )

            result = Signal(
                np.append(self.samples, other.samples, axis=0),
                np.append(self.timestamps, timestamps),
                self.unit,
                self.name,
                self.conversion,
                self.comment,
                self.raw,
                self.master_metadata,
                self.display_names,
                self.attachment,
                self.source,
                self.bit_count,
                invalidation_bits=invalidation_bits,
                encoding=self.encoding,
                group_index=self.group_index,
                channel_index=self.channel_index,
                flags=self.flags,
                virtual_conversion=self.virtual_conversion,
                virtual_master_conversion=self.virtual_master_conversion,
            )
        else:
            result = self

        return result

    def interp(
        self,
        new_timestamps: NDArray[Any] | list[float],
        integer_interpolation_mode: (
            IntInterpolationModeType | IntegerInterpolation
        ) = IntegerInterpolation.REPEAT_PREVIOUS_SAMPLE,
        float_interpolation_mode: (
            FloatInterpolationModeType | FloatInterpolation
        ) = FloatInterpolation.LINEAR_INTERPOLATION,
    ) -> "Signal":
        """Return a new `Signal` interpolated using the `new_timestamps`.

        Parameters
        ----------
        new_timestamps : np.ndarray | list
            Timestamps used for interpolation.

        integer_interpolation_mode : int, default 0
            Interpolation mode for integer signals.

            * 0 - repeat previous sample
            * 1 - linear interpolation
            * 2 - hybrid interpolation: channels with integer data type (raw
              values) that have a conversion that outputs float values will use
              linear interpolation, otherwise the previous sample is used

            .. versionadded:: 6.2.0

        float_interpolation_mode : int, default 1
            Interpolation mode for float channels.

            * 0 - repeat previous sample
            * 1 - linear interpolation

            .. versionadded:: 6.2.0

        Returns
        -------
        signal : Signal
            New interpolated `Signal`.
        """

        integer_interpolation_mode = IntegerInterpolation(integer_interpolation_mode)
        float_interpolation_mode = FloatInterpolation(float_interpolation_mode)

        if not len(self.samples) or not len(new_timestamps):
            return Signal(
                self.samples[:0].copy(),
                self.timestamps[:0].copy(),
                self.unit,
                self.name,
                comment=self.comment,
                conversion=self.conversion,
                raw=self.raw,
                master_metadata=self.master_metadata,
                display_names=self.display_names,
                attachment=self.attachment,
                invalidation_bits=None,
                encoding=self.encoding,
                group_index=self.group_index,
                channel_index=self.channel_index,
                flags=self.flags,
                virtual_conversion=self.virtual_conversion,
                virtual_master_conversion=self.virtual_master_conversion,
            )
        else:
            # # we need to validate first otherwise we can get false invalid data
            # # if the new timebase and the invalidation bits are aligned in an
            # # infavorable way
            #
            # if self.invalidation_bits is not None:
            #     signal = self.validate()
            #     has_invalidation = True
            # else:
            #     signal = self
            #     has_invalidation = False

            signal = self
            invalidation_bits: NDArray[np.bool] | None = signal.invalidation_bits

            if not len(signal.samples):
                return Signal(
                    self.samples[:0].copy(),
                    self.timestamps[:0].copy(),
                    self.unit,
                    self.name,
                    comment=self.comment,
                    conversion=self.conversion,
                    raw=self.raw,
                    master_metadata=self.master_metadata,
                    display_names=self.display_names,
                    attachment=self.attachment,
                    invalidation_bits=None if invalidation_bits is None else np.array([], dtype=bool),
                    encoding=self.encoding,
                    group_index=self.group_index,
                    channel_index=self.channel_index,
                    flags=self.flags,
                    virtual_conversion=self.virtual_conversion,
                    virtual_master_conversion=self.virtual_master_conversion,
                )

            if len(signal.samples.shape) > 1:
                idx = np.searchsorted(signal.timestamps, new_timestamps, side="right")
                idx -= 1
                idx[idx < 0] = 0
                s = signal.samples[idx]
                if invalidation_bits is not None:
                    invalidation_bits = invalidation_bits[idx]
            else:
                kind = signal.samples.dtype.kind

                if kind == "f":
                    if float_interpolation_mode == FloatInterpolation.REPEAT_PREVIOUS_SAMPLE:
                        idx = np.searchsorted(signal.timestamps, new_timestamps, side="right")
                        idx -= 1
                        idx[idx < 0] = 0
                        s = signal.samples[idx]

                        if invalidation_bits is not None:
                            invalidation_bits = invalidation_bits[idx]

                    else:
                        s = np.interp(new_timestamps, signal.timestamps, signal.samples)

                        if invalidation_bits is not None:
                            idx = np.searchsorted(signal.timestamps, new_timestamps, side="right")
                            idx -= 1
                            idx[idx < 0] = 0
                            invalidation_bits = invalidation_bits[idx]

                elif kind in "ui":
                    if integer_interpolation_mode == IntegerInterpolation.HYBRID_INTERPOLATION:
                        if signal.raw and signal.conversion:
                            kind = signal.conversion.convert(signal.samples[:1]).dtype.kind
                            if kind == "f":
                                integer_interpolation_mode = IntegerInterpolation.LINEAR_INTERPOLATION

                    if integer_interpolation_mode == IntegerInterpolation.HYBRID_INTERPOLATION:
                        integer_interpolation_mode = IntegerInterpolation.REPEAT_PREVIOUS_SAMPLE

                    if integer_interpolation_mode == IntegerInterpolation.LINEAR_INTERPOLATION:
                        s = np.interp(new_timestamps, signal.timestamps, signal.samples).astype(signal.samples.dtype)

                        if invalidation_bits is not None:
                            idx = np.searchsorted(signal.timestamps, new_timestamps, side="right")
                            idx -= 1
                            idx[idx < 0] = 0
                            invalidation_bits = invalidation_bits[idx]

                    elif integer_interpolation_mode == IntegerInterpolation.REPEAT_PREVIOUS_SAMPLE:
                        idx = np.searchsorted(signal.timestamps, new_timestamps, side="right")
                        idx -= 1
                        idx[idx < 0] = 0

                        s = signal.samples[idx]

                        if invalidation_bits is not None:
                            invalidation_bits = invalidation_bits[idx]

                else:
                    idx = np.searchsorted(signal.timestamps, new_timestamps, side="right")
                    idx -= 1
                    idx[idx < 0] = 0
                    s = signal.samples[idx]

                    if invalidation_bits is not None:
                        invalidation_bits = invalidation_bits[idx]

            if s.dtype != self.samples.dtype:
                s = s.astype(self.samples.dtype)

            return Signal(
                s,
                new_timestamps,
                self.unit,
                self.name,
                comment=self.comment,
                conversion=self.conversion,
                source=self.source,
                raw=self.raw,
                master_metadata=self.master_metadata,
                display_names=self.display_names,
                attachment=self.attachment,
                invalidation_bits=invalidation_bits,
                encoding=self.encoding,
                group_index=self.group_index,
                channel_index=self.channel_index,
                flags=self.flags,
                virtual_conversion=self.virtual_conversion,
                virtual_master_conversion=self.virtual_master_conversion,
            )

    def __apply_func(self, other: Union["Signal", NDArray[Any], float] | None, func_name: str) -> "Signal":
        """Delegate operations to the `samples` attribute, but in a
        time-correct manner by considering the `timestamps`.
        """

        if isinstance(other, Signal):
            if len(self) and len(other):
                start = max(self.timestamps[0], other.timestamps[0])
                stop = min(self.timestamps[-1], other.timestamps[-1])
                s1 = self.physical().cut(start, stop)
                s2 = other.physical().cut(start, stop)
            else:
                s1 = self
                s2 = other

            time = np.union1d(s1.timestamps, s2.timestamps)
            s1 = s1.interp(time)
            s2 = s2.interp(time)

            invalidation_bits: NDArray[np.bool] | None
            if s1.invalidation_bits is not None or s2.invalidation_bits is not None:
                if s1.invalidation_bits is None:
                    invalidation_bits = s2.invalidation_bits
                elif s2.invalidation_bits is None:
                    invalidation_bits = s1.invalidation_bits
                else:
                    invalidation_bits = s1.invalidation_bits | s2.invalidation_bits
            else:
                invalidation_bits = None

            func = getattr(s1.samples, func_name)
            conversion = None
            s = func(s2.samples)

        elif other is None:
            s = self.samples
            conversion = self.conversion
            time = self.timestamps
            invalidation_bits = self.invalidation_bits

        else:
            func = getattr(self.samples, func_name)
            s = func(other)
            conversion = self.conversion
            time = self.timestamps
            invalidation_bits = self.invalidation_bits

        return Signal(
            samples=s,
            timestamps=time,
            unit=self.unit,
            name=self.name,
            conversion=conversion,
            raw=self.raw,
            master_metadata=self.master_metadata,
            display_names=self.display_names,
            attachment=self.attachment,
            invalidation_bits=invalidation_bits,
            source=self.source,
            encoding=self.encoding,
            group_index=self.group_index,
            channel_index=self.channel_index,
            flags=self.flags,
            virtual_conversion=self.virtual_conversion,
            virtual_master_conversion=self.virtual_master_conversion,
        )

    def __pos__(self) -> "Signal":
        return self

    def __neg__(self) -> "Signal":
        return Signal(
            np.negative(self.samples),
            self.timestamps,
            unit=self.unit,
            name=self.name,
            conversion=self.conversion,
            raw=self.raw,
            master_metadata=self.master_metadata,
            display_names=self.display_names,
            attachment=self.attachment,
            invalidation_bits=self.invalidation_bits,
            source=self.source,
            encoding=self.encoding,
            flags=self.flags,
            virtual_conversion=self.virtual_conversion,
            virtual_master_conversion=self.virtual_master_conversion,
        )

    def __round__(self, n: int) -> "Signal":
        return Signal(
            np.around(self.samples, n),
            self.timestamps,
            unit=self.unit,
            name=self.name,
            conversion=self.conversion,
            raw=self.raw,
            master_metadata=self.master_metadata,
            display_names=self.display_names,
            attachment=self.attachment,
            invalidation_bits=self.invalidation_bits,
            source=self.source,
            encoding=self.encoding,
            flags=self.flags,
            virtual_conversion=self.virtual_conversion,
            virtual_master_conversion=self.virtual_master_conversion,
        )

    def __sub__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__sub__")

    def __isub__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__sub__(other)

    def __rsub__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return -self.__sub__(other)

    def __add__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__add__")

    def __iadd__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__add__(other)

    def __radd__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__add__(other)

    def __truediv__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__truediv__")

    def __itruediv__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__truediv__(other)

    def __rtruediv__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__rtruediv__")

    def __mul__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__mul__")

    def __imul__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__mul__(other)

    def __rmul__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__mul__(other)

    def __floordiv__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__floordiv__")

    def __ifloordiv__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__truediv__(other)

    def __rfloordiv__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return 1 / self.__apply_func(other, "__rfloordiv__")

    def __mod__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__mod__")

    def __pow__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__pow__")

    def __and__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__and__")

    def __or__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__or__")

    def __xor__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__xor__")

    def __invert__(self) -> "Signal":
        s = ~self.samples
        time = self.timestamps
        return Signal(
            s,
            time,
            unit=self.unit,
            name=self.name,
            conversion=self.conversion,
            raw=self.raw,
            master_metadata=self.master_metadata,
            display_names=self.display_names,
            attachment=self.attachment,
            invalidation_bits=self.invalidation_bits,
            source=self.source,
            encoding=self.encoding,
            flags=self.flags,
            virtual_conversion=self.virtual_conversion,
            virtual_master_conversion=self.virtual_master_conversion,
        )

    def __lshift__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__lshift__")

    def __rshift__(self, other: Union["Signal", NDArray[Any], float] | None) -> "Signal":
        return self.__apply_func(other, "__rshift__")

    def __lt__(self, other: Union["Signal", NDArray[Any]] | None) -> "Signal":
        return self.__apply_func(other, "__lt__")

    def __le__(self, other: Union["Signal", NDArray[Any]] | None) -> "Signal":
        return self.__apply_func(other, "__le__")

    def __gt__(self, other: Union["Signal", NDArray[Any]] | None) -> "Signal":
        return self.__apply_func(other, "__gt__")

    def __ge__(self, other: Union["Signal", NDArray[Any]] | None) -> "Signal":
        return self.__apply_func(other, "__ge__")

    def __eq__(self, other: Union["Signal", NDArray[Any]] | None) -> "Signal":  # type: ignore[override]
        return self.__apply_func(other, "__eq__")

    def __ne__(self, other: Union["Signal", NDArray[Any]] | None) -> "Signal":  # type: ignore[override]
        return self.__apply_func(other, "__ne__")

    def __iter__(self) -> Iterator[NDArray[Any] | str]:
        yield from (self.samples, self.timestamps, self.unit, self.name)

    def __reversed__(self) -> Iterator[tuple[int, tuple[Any, Any]]]:
        return enumerate(zip(reversed(self.samples), reversed(self.timestamps), strict=False))

    def __len__(self) -> int:
        return len(self.samples)

    def __abs__(self) -> "Signal":
        return Signal(
            np.fabs(self.samples),
            self.timestamps,
            unit=self.unit,
            name=self.name,
            conversion=self.conversion,
            raw=self.raw,
            master_metadata=self.master_metadata,
            display_names=self.display_names,
            attachment=self.attachment,
            invalidation_bits=self.invalidation_bits,
            source=self.source,
            flags=self.flags,
            virtual_conversion=self.virtual_conversion,
            virtual_master_conversion=self.virtual_master_conversion,
        )

    @overload
    def __getitem__(self, val: str) -> NDArray[Any]: ...

    @overload
    def __getitem__(self, val: int | slice) -> "Signal": ...

    def __getitem__(self, val: int | slice | str) -> Union[NDArray[Any], "Signal"]:
        if isinstance(val, str):
            return self.samples[val]
        else:
            return Signal(
                self.samples[val],
                self.timestamps[val],
                self.unit,
                self.name,
                self.conversion,
                self.comment,
                self.raw,
                self.master_metadata,
                self.display_names,
                self.attachment,
                self.source,
                self.bit_count,
                invalidation_bits=self.invalidation_bits[val] if self.invalidation_bits is not None else None,
                encoding=self.encoding,
                group_index=self.group_index,
                channel_index=self.channel_index,
                flags=self.flags,
                virtual_conversion=self.virtual_conversion,
                virtual_master_conversion=self.virtual_master_conversion,
            )

    def __setitem__(self, idx: Any, val: Any) -> None:
        self.samples[idx] = val

    def astype(self, np_type: DTypeLike) -> "Signal":
        """Return a new `Signal` with samples of dtype `np_type`.

        Parameters
        ----------
        np_type : np.dtype
            New numpy dtype.

        Returns
        -------
        signal : Signal
            New `Signal` with the samples of dtype `np_type`.
        """
        return Signal(
            self.samples.astype(np_type),
            self.timestamps,
            unit=self.unit,
            name=self.name,
            conversion=self.conversion,
            raw=self.raw,
            master_metadata=self.master_metadata,
            display_names=self.display_names,
            attachment=self.attachment,
            invalidation_bits=self.invalidation_bits,
            source=self.source,
            encoding=self.encoding,
            flags=self.flags,
            virtual_conversion=self.virtual_conversion,
            virtual_master_conversion=self.virtual_master_conversion,
        )

    def physical(self, copy: bool = True, ignore_value2text_conversions: bool = False) -> "Signal":
        """Get the physical sample values.

        Parameters
        ----------
        copy : bool, default True
            Copy the samples and timestamps in the returned Signal.

            .. versionadded:: 7.4.0

        ignore_value2text_conversions : bool, default False
            Make sure that the output signal has numeric samples by ignoring
            the value to text conversions.

            .. versionadded:: 8.3.0

        Returns
        -------
        phys : Signal
            New `Signal` with physical values.
        """

        if not self.raw or self.conversion is None:
            if copy:
                samples = self.samples.copy()
            else:
                samples = self.samples
            encoding = None
        else:
            samples = self.conversion.convert(self.samples, ignore_value2text_conversions=ignore_value2text_conversions)
            if samples.dtype.kind == "S":
                encoding = "utf-8" if self.conversion.id == b"##CC" else "latin-1"
            else:
                encoding = None

        return Signal(
            samples,
            self.timestamps.copy() if copy else self.timestamps,
            unit=self.unit,
            name=self.name,
            conversion=None,
            raw=False,
            master_metadata=self.master_metadata,
            display_names=self.display_names,
            attachment=self.attachment,
            invalidation_bits=self.invalidation_bits,
            source=self.source,
            encoding=encoding,
            group_index=self.group_index,
            channel_index=self.channel_index,
            flags=self.flags,
            virtual_conversion=self.virtual_conversion,
            virtual_master_conversion=self.virtual_master_conversion,
        )

    def validate(self, copy: bool = True) -> "Signal":
        """Apply invalidation bits if they are available for this signal.

        Parameters
        ----------
        copy : bool, default True
            Return a copy of the result.

            .. versionadded:: 5.12.0
        """
        if self.invalidation_bits is None:
            signal = self

        else:
            idx = np.nonzero(~self.invalidation_bits)[0]
            if len(idx) == len(self.samples):
                signal = self
            else:
                signal = Signal(
                    self.samples[idx],
                    self.timestamps[idx],
                    self.unit,
                    self.name,
                    self.conversion,
                    self.comment,
                    self.raw,
                    self.master_metadata,
                    self.display_names,
                    self.attachment,
                    self.source,
                    self.bit_count,
                    invalidation_bits=None,
                    encoding=self.encoding,
                    group_index=self.group_index,
                    channel_index=self.channel_index,
                    flags=self.flags,
                    virtual_conversion=self.virtual_conversion,
                    virtual_master_conversion=self.virtual_master_conversion,
                )

        if copy:
            signal = signal.copy()

        return signal

    def copy(self) -> "Signal":
        """Copy all attributes to a new Signal."""
        return Signal(
            self.samples.copy(),
            self.timestamps.copy(),
            self.unit,
            self.name,
            self.conversion,
            self.comment,
            self.raw,
            self.master_metadata,
            self.display_names,
            self.attachment,
            self.source,
            self.bit_count,
            invalidation_bits=self.invalidation_bits.copy() if self.invalidation_bits is not None else None,
            encoding=self.encoding,
            group_index=self.group_index,
            channel_index=self.channel_index,
            flags=self.flags,
            virtual_conversion=self.virtual_conversion,
            virtual_master_conversion=self.virtual_master_conversion,
        )


if __name__ == "__main__":
    pass
